#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <dirent.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/vt.h>
#include <linux/fb.h>
#include <linux/kd.h>

#include "ndrver.h"

#include "MoreAppleDefs.h"
#include "IOMacOSVideo.h"

#define NDRVS_PATH	"/etc/ndrver"

struct ndrv_info
{
	char			name[256];	/* Driver name */
	char			match[256];	/* Device matching name */
	unsigned int		version;	/* Driver version */
	unsigned int		crdate;		/* Creation date */
};

struct ndrv_file_spec
{
	char			*path;		/* File name */
	unsigned long		offset;		/* Offset of PEF within file */
	unsigned long		size;		/* Size of PEF within file */
	struct ndrv_info	info;		/* Parsed PEF data */
	struct ndrv_file_spec	*next;
};

struct ndrv_file_spec *ndrv_file_specs;
struct ndrv_inst *ndrvs;
static FILE *log_file;

void p2cstr(char *cdest, unsigned char *psrc)
{
	int len = psrc[0];
	memmove(cdest, psrc + 1, len);
	cdest[len] = 0;
}

void log_printf(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vprintf(fmt, args);
	va_end(args);
	va_start(args, fmt);
	vfprintf(log_file, fmt, args);
	va_end(args);
}

void IOLog(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	log_printf(fmt, args);
	va_end(args);
}

static const IOTVector * undef_sym_handler(void * self, const char * libraryName, const char * symbolName)
{
	printf("** undefined symbol %s in lib: %s\n", symbolName, libraryName);
	
	return NULL;
}

void libcall_trace(struct FunctionEntry *entry, void *caller, unsigned long args[])
{
	printf("[LIB] LibCall: %s from %p\n", entry->name, caller);
}

static int print_vers(char *str, unsigned int vers)
{
	char *p = str;
	p += sprintf(p, "%d.%d.%d",
		     (int)(((vers >> 28) & 0xf) * 10 + ((vers >> 24) & 0xf)),
		     (int)(vers >> 20) & 0xf,
		     (int)(vers >> 16) & 0xf);
	switch((vers >> 12) & 0xf) {
	case 0x2: p += sprintf(p, "d"); break;
	case 0x4: p += sprintf(p, "a"); break;
	case 0x6: p += sprintf(p, "b"); break;
	case 0x8: p += sprintf(p, "f"); break;
	default : p += sprintf(p, "?"); break;
	}
	p += sprintf(p, "%d", (int)vers & 0xff);

	return p - str;
}

static void * load_file(const char *path, long *size)
{
	int fd;
	void *buf;

	fd = open(path, 0);
	if (fd == -1) {
		perror("can't open file\n");
		return NULL;
	}
	*size = lseek(fd, 0, SEEK_END);
	lseek(fd, 0, SEEK_SET);

	buf = malloc(*size);
	if (buf == NULL) {
		perror("can't allocate memory\n");
		close(fd);
		return NULL;
	}

	read(fd, buf, *size);
	close(fd);

	return buf;
}

static void * load_file_chunk(const char *path, long offset, long *size)
{
	int fd;
	void *buf;
	long fsize;

	fd = open(path, 0);
	if (fd == -1) {
		perror("can't open file\n");
		return NULL;
	}
	fsize = lseek(fd, 0, SEEK_END);
	if (offset >= fsize) {
		close(fd);
		return NULL;
	}
	if ((fsize - offset) < *size)
		*size = fsize - offset;
	lseek(fd, offset, SEEK_SET);

	buf = malloc(*size);
	if (buf == NULL) {
		perror("can't allocate memory\n");
		close(fd);
		return NULL;
	}

	read(fd, buf, *size);
	close(fd);

	return buf;
}

static int get_ndrv_info(void *pefndrv, long *pefsize, struct ndrv_info *info)
{
	PCodeInstance pinst;
	struct IOTVector *sDoDriverIO;
	struct DriverDescription *sDriverDescription;
	UInt32 size, date;
	int rc = 0;

	memset(info, 0, sizeof(struct ndrv_info));

        rc = PCodeOpen(pefndrv, *pefsize, &pinst, &date, &size);
	if (rc) {
		fprintf(stderr, "PCodeOpen(): returned %d\n", rc);
		return -1;
	}
	*pefsize = size;

        rc = PCodeInstantiate(pinst, undef_sym_handler, 0);
	if (rc) {
		fprintf(stderr, "PCodeInstanciate(): returned %d\n", rc);
		goto bail;
	}

	/* Check that DoDriverIO exists */
	rc = PCodeFindExport(pinst, "DoDriverIO", (IOLogicalAddress *)&sDoDriverIO, NULL);
	if (rc) {
		fprintf(stderr, "Can't find symbol DoDriverIO !\n");
		goto bail;
	}
	rc = PCodeFindExport(pinst, "TheDriverDescription", (IOLogicalAddress *)&sDriverDescription, NULL);
	if (rc) {
		fprintf(stderr, "Can't find symbol TheDriverDescription !\n");
		goto bail;
	}

	p2cstr(info->name, sDriverDescription->driverOSRuntimeInfo.driverName);
	p2cstr(info->match, sDriverDescription->driverType.nameInfoStr);
	info->version = *(UInt32 *)&sDriverDescription->driverType.version;
	info->crdate = date;

 bail:
	PCodeClose(pinst);

	return rc;
}

/*
 * Parse all file-based ndrvs and build a list of them with their versions
 * used for matching against device-tree entries 
 */
static void load_file_ndrvs(void)
{
	char *curpath;
	DIR *drvdir;
	struct dirent *dent;

	/* Get to /etc/ndrver */
	drvdir = opendir(NDRVS_PATH);
	if (drvdir == NULL)
		return;
	
	/* Try each file in there */
	while((dent = readdir(drvdir)) != NULL) {
		long cursor = 0;
		int ndrvcount = 0;
		void *filedata;
		long filesize;

		if (dent->d_type == DT_DIR)
			continue;
		curpath = malloc(strlen(NDRVS_PATH) + strlen(dent->d_name) + 2);
		sprintf(curpath, "%s/%s", NDRVS_PATH, dent->d_name);
		LOG_INFO("Loading driver file: %s\n", curpath);
		filedata = load_file(curpath, &filesize);
		if (filedata == NULL || filesize < 16) {
			LOG_ERR("Load failed\n");
			goto nextfile;
		}
		LOG_DBG("File size: %ld bytes\n", filesize);

		/* Check it's really a PEF */
		if (*((UInt32 *)(filedata + cursor)) != 'Joy!' ||
		    *((UInt32 *)(filedata + cursor + 4)) != 'peff') {
			LOG_ERR("Not a PEF file !\n");
			goto nextfile;
		}
		LOG_DBG("Looking for drivers in file...\n");

		/* Find all drivers in there */
		while((((long)filesize) - cursor) > 16) {
			struct ndrv_file_spec *spec = malloc(sizeof(struct ndrv_file_spec));
			long sz;
			int rc;

			if (spec == NULL) {
				LOG_ERR("Out of memory !\n");
				goto nextfile;
			}
			/* Parse the sucker */
			sz = filesize - cursor;
			rc = get_ndrv_info(filedata + cursor, &sz, &spec->info);
			if (rc != 0) {
				LOG_ERR("Failed to get NDRV infos at offset 0x%x\n",
				    (unsigned int)cursor);
				free(spec);
				goto nextfile;
			}
			LOG_INFO("Found driver for: %s\n", spec->info.match);

			/* Fill out remaining data */
			ndrvcount++;
			spec->path = curpath;
			spec->offset = cursor;
			spec->size = sz;
			spec->next = ndrv_file_specs;
			ndrv_file_specs = spec;
			
			/* Try to find next one */
			cursor += sz;
			cursor = (cursor + 0x3) & ~3ul;
			while((filesize - cursor) > 16 &&
			      *((UInt32 *)(filedata + cursor)) != 'Joy!')
				cursor += 4;
		}
	nextfile:
		if (ndrvcount == 0)
			free(curpath);
		LOG_INFO("Found %d NDRV's in file\n", ndrvcount);
		if (filedata)
			free(filedata);
	}
	closedir(drvdir);
}

static int instanciate_raw_ndrv(struct ndrv_inst *ndrv)
{
	int rc = 0;

        rc = PCodeOpen(ndrv->rawpef, ndrv->rawsize, &ndrv->pinst, NULL, NULL);
	if (rc) {
		fprintf(stderr, "PCodeOpen(): returned %d\n", rc);
		return -1;
	}

        rc = PCodeInstantiate(ndrv->pinst, undef_sym_handler, 0);
	if (rc) {
		fprintf(stderr, "PCodeInstanciate(): returned %d\n", rc);
		goto bail;
	}

	/* Check that DoDriverIO exists */
	rc = PCodeFindExport(ndrv->pinst, "DoDriverIO",
			     (IOLogicalAddress *)&ndrv->sDoDriverIO, NULL);
	if (rc) {
		fprintf(stderr, "Can't find symbol DoDriverIO !\n");
		goto bail;
	}
	rc = PCodeFindExport(ndrv->pinst, "TheDriverDescription",
			     (IOLogicalAddress *)&ndrv->sDriverDescription, NULL);
	if (rc) {
		fprintf(stderr, "Can't find symbol TheDriverDescription !\n");
		goto bail;
	}

 bail:
	if (rc != 0)
		PCodeClose(ndrv->pinst);

	return rc;
}


static struct ndrv_inst *instanciate_rom_ndrv(struct device_node *np,
					      void *pef, long pefsize)
{
	return 0;
}

static struct ndrv_inst *instanciate_file_ndrv(struct device_node *np,
					       struct ndrv_file_spec *spec)
{
	struct ndrv_inst *ndrv;
	int rc;

	LOG_INFO("Instanciating file based ndrv...\n");

	/* Allocate instance structure */
	ndrv = malloc(sizeof(struct ndrv_inst));
	if (ndrv == NULL)
		return NULL;
	memset(ndrv, 0, sizeof(struct ndrv_inst));
	ndrv->node = np;

	/* Do the matching with the PCI layer */
	rc = ndrvpci_init_instance(ndrv);
	if (rc != 0) {
		LOG_ERR("Matching of node to PCI failed !\n");
		return NULL;
	}

	/* Load file */
	ndrv->rawsize = spec->size;
	ndrv->rawpef = load_file_chunk(spec->path, spec->offset, &ndrv->rawsize);
	if (ndrv->rawpef == NULL) {
		LOG_ERR("Failed loading from disk !\n");
		free(ndrv);
		return NULL;
	}
	
	/* Instanciate the PEF itself */
	rc = instanciate_raw_ndrv(ndrv);
	if (rc != 0) {
		LOG_ERR("PEF instanciation failed !\n");
		free(ndrv->rawpef);
		free(ndrv);
		return NULL;
	}
	
	return ndrv;;
}

/* Magic date from Darwin: do not use a ROM driver older than that */
#define kDate2001March1 0xb6c49300

/* Match a "display" device to an ndrv */
static struct ndrv_inst *match_display_node(struct device_node *np)
{
	struct ndrv_info rompef_info;
	void *rompef;
	long rompefsize;
	struct ndrv_file_spec *ovspec;
	char vstring[32];
	struct ndrv_inst *ndrv = NULL;
	int temp;

	LOG_INFO("Matching display: %s\n", np->full_name);

	/* First check for ROM based ndrv */
	rompef = get_property(np, "driver,AAPL,MacOS,PowerPC", &temp);
	if (rompef) {
		rompefsize = temp;
		if (get_ndrv_info(rompef, &rompefsize, &rompef_info) != 0)
			rompef = NULL;
	}
	/* Check date of ROM based ndrv */
	if (rompef && rompef_info.crdate < kDate2001March1) {
		rompef = NULL;
		LOG_INFO("ROM driver out of date !\n");
	}
	if (rompef) {
		print_vers(vstring, rompef_info.version);
		LOG_INFO("Found ROM driver, version: %s\n", vstring);
	}

	/* Now check for override file based driver */
	/* XXX maybe it would be wise to scan for the highest version number in case
	 * a driver appears twice
	 */
	for (ovspec = ndrv_file_specs; ovspec != NULL; ovspec = ovspec->next) {
		if (strcmp(ovspec->info.match, np->name) && !
		    device_is_compatible(np, ovspec->info.match))
			continue;
		print_vers(vstring, ovspec->info.version);
		LOG_INFO("Found candidate file driver version: %s (match: %s)\n",
		    vstring, ovspec->info.match);
		break;
	}
	
	/* Compare file vs. ROM driver versions */ 
	if (ovspec && rompef && ovspec->info.version > rompef_info.version) {
		rompef = NULL;
		LOG_INFO("File driver overrides ROM driver\n");
	}

	/* Instanciate */
	/* XXX Should avoid duplicating the driver PEF in memory
	 */
	if (rompef)
		ndrv = instanciate_rom_ndrv(np, rompef, rompefsize);
	if (ndrv == NULL && ovspec)
		ndrv = instanciate_file_ndrv(np, ovspec);
	if (ndrv) {
		ndrv->next = ndrvs;
		ndrvs = ndrv;
	}
	return ndrv;
}

/*
 * Entry point to NDRV
 */
static OSStatus do_driver_io(struct ndrv_inst *ndrv,
			     UInt32 commandID, void *commandParams,
			     UInt32 commandCode, UInt32 commandKind)
{
        return CallTVector(0, (void *)commandID, commandParams,
			   (void *)commandCode, (void *)commandKind,
			   0, ndrv->sDoDriverIO);
}

static OSStatus do_status(struct ndrv_inst *ndrv, int csCode, void *params)
{
	struct CntrlParam pb;

	memset(&pb, 0, sizeof(pb));
	pb.qLink = 0;
	pb.csCode = csCode;
	pb.csParams = params;

	return do_driver_io(ndrv, 0, &pb, kStatusCommand, kImmediateIOCommandKind);
}


static OSStatus do_control(struct ndrv_inst *ndrv, int csCode, void *params)
{
	struct CntrlParam pb;

	memset(&pb, 0, sizeof(pb));
	pb.qLink = 0;
	pb.csCode = csCode;
	pb.csParams = params;

	return do_driver_io(ndrv, 0, &pb, kControlCommand, kImmediateIOCommandKind);
}

/*
 * Do stuffs....
 */
static void ndrv_initialize(struct ndrv_inst *ndrv)
{
	struct DriverInitInfo initInfo;
	UInt32 value;
	OSStatus err;
	
	value = 1;
	set_property(ndrv->node, "AAPL,gray-page", &value, 4);

	value = 0x00555555;
	set_property(ndrv->node, "AAPL,gray-value", &value, 4);

	LOG_INFO("Calling Initialize ...\n");
	memset(&initInfo, 0, sizeof(initInfo));
	initInfo.refNum = -10;
	SET_REGENTRY_TO_DN(&initInfo.deviceEntry, ndrv->node);
	err = do_driver_io(ndrv, 0, &initInfo, kInitializeCommand,
			   kImmediateIOCommandKind);
	LOG_INFO("Initialize complete, rc: %ld !\n", err);	
}

static void ndrv_open(struct ndrv_inst *ndrv)
{
	struct CntrlParam pb;
	OSStatus err;
	VDResolutionInfoRec resolution;

	LOG_INFO("Calling Open for %s ...\n", ndrv->node->full_name);	
	memset(&pb, 0, sizeof(pb));
	pb.ioCRefNum = 0xffff & ((UInt32) ndrv);
	err = do_driver_io(ndrv, 0, &pb, kOpenCommand, kImmediateIOCommandKind);
	LOG_INFO("Open complete, rc: %ld !\n", err);

	resolution.csDisplayModeID = kDisplayModeIDInvalid;
}

static void do_ndrv_openall(void)
{
	struct ndrv_inst *ndrv;

	for (ndrv = ndrvs; ndrv; ndrv = ndrv->next)
		ndrv_open(ndrv);
	
}

static void ndrv_suspend(struct ndrv_inst *ndrv)
{
	OSStatus err;
        VDPowerStateRec	sleepInfo;

	LOG_INFO("Suspend: Getting power state for %s ...\n", ndrv->node->full_name);	
	memset(&sleepInfo, 0, sizeof(sleepInfo));
	err = do_status(ndrv, cscGetPowerState, &sleepInfo);
	if (err != 0) {
		LOG_ERR("cscGetPowerState failed !\n");
		return;
	}
	LOG_INFO("powerFlags: %lx, powerState: %lx\n", 
		 sleepInfo.powerFlags, sleepInfo.powerState);
	sleepInfo.powerState = kHardwareSleep;
	sleepInfo.powerFlags = 0;
	sleepInfo.powerReserved1 = kPowerStateSleepWakeNeedsProbeMask;
	sleepInfo.powerReserved2 = 0;
	err = do_control(ndrv, cscSetPowerState, &sleepInfo);
	LOG_INFO("Suspend complete, rc: %ld !\n", err);

	LOG_INFO("Suspend: Saving PCI for %s ...\n", ndrv->node->full_name);	
	ndrvpci_save_config(ndrv);
}

static void do_ndrv_suspendall(void)
{
	struct ndrv_inst *ndrv;

	/* Suspend kernel drivers using the special sleep mode */
	ndrvpci_kernel_suspendall();

	/* Call NDRV's suspend */
	for (ndrv = ndrvs; ndrv; ndrv = ndrv->next)
		ndrv_suspend(ndrv);
	
}

static void ndrv_resume(struct ndrv_inst *ndrv)
{
	OSStatus err;
        VDPowerStateRec	sleepInfo;

	LOG_INFO("Resume: Restoring PCI for %s ...\n", ndrv->node->full_name);	
	ndrvpci_restore_config(ndrv);

	LOG_INFO("Resume: Getting power state for %s ...\n", ndrv->node->full_name);	
	memset(&sleepInfo, 0, sizeof(sleepInfo));
	err = do_status(ndrv, cscGetPowerState, &sleepInfo);
	if (err != 0) {
		LOG_ERR("cscGetPowerState failed !\n");
		return;
	}
	LOG_INFO("powerFlags: %lx, powerState: %lx\n", 
		 sleepInfo.powerFlags, sleepInfo.powerState);
	sleepInfo.powerState = kHardwareWake;
	sleepInfo.powerFlags = 0;
	sleepInfo.powerReserved1 = kPowerStateSleepWakeNeedsProbeMask;
	sleepInfo.powerReserved2 = 0;
	err = do_control(ndrv, cscSetPowerState, &sleepInfo);
	LOG_INFO("Resume complete, rc: %ld !\n", err);
}

static void do_ndrv_resumeall(void)
{
	struct ndrv_inst *ndrv;

	/* Call NDRV's resume */
	for (ndrv = ndrvs; ndrv; ndrv = ndrv->next)
		ndrv_resume(ndrv);
	
	/* Resume kernel drivers using the special wake mode */
	ndrvpci_kernel_resumeall();
}


/* Scan device-tree for "display" type devices
 */
static void do_ndrv_matching(void)
{
	struct device_node *np;
	struct ndrv_inst *ndrv;
		
	for (np = allnodes ; np != NULL ; np = np->allnext) {
		char *type = get_property(np, "device_type", NULL);
		if (type == NULL)
			continue;
		if (strcmp(type, "display"))
			continue;

		/* We have a candidate */
		ndrv = match_display_node(np);
		if (ndrv)
			ndrv_initialize(ndrv);
	}
}

extern void InitLinuxR2(void);

static int vt_orig;
static int vt_no;
static int console_fd = -1;

static void restore_vt(void)
{
	if (console_fd == -1)
		return;
	ioctl( console_fd, KDSETMODE, KD_TEXT );
	ioctl( console_fd, VT_ACTIVATE, vt_orig );
	ioctl( console_fd, VT_WAITACTIVE, vt_orig );
	close( console_fd );
	console_fd = -1;
}

static int setup_vt(void)
{
	struct vt_stat 	vts;
	int 		fd;
	char 		vtname[32];
	struct termios 	tio;	

	if ((fd = open("/dev/tty0", O_WRONLY)) < 0) {
		fprintf(stderr, "failed to open /dev/tty0: %s\n", strerror(errno));
		return 1;
	}

	/* find original vt */
	if (ioctl(fd, VT_GETSTATE, &vts) < 0)
		return 1;
	vt_orig = vts.v_active;

	/* find free console? */
	if ((ioctl(fd, VT_OPENQRY, &vt_no) < 0) || (vt_no == -1)) {
		fprintf(stderr, "Cannot find a free VT\n");
		return 1;
	}
	close(fd);

	/* detach old tty */
	setpgrp();
       	if ((fd = open("/dev/tty",O_RDWR)) >= 0) {
		ioctl(fd, TIOCNOTTY, 0);
		close(fd);
	}

	/* open our vt */
	sprintf(vtname,"/dev/tty%d",vt_no);
	if ((console_fd = open(vtname, O_RDWR )) < 0) {
		fprintf(stderr, "Cannot open %s (%s)\n",vtname, strerror(errno));
		return 1;
	}

#if 0
	/* set controlling tty */
	if( ioctl( console_fd, TIOCSCTTY, 1 ) < 0 ){
		perror("Error setting controlling tty");
		return 1; 
	}
#endif

	/* No line buffering or echoing */
	tcgetattr( console_fd, &tio );
	tio.c_lflag &= ~(ICANON | ECHO);
	tcsetattr( console_fd, TCSANOW, &tio );

	/* Switch VT */
	ioctl( console_fd, VT_ACTIVATE, vt_no );
	ioctl( console_fd, VT_WAITACTIVE, vt_no );

	/* Set video mode  */
	ioctl( console_fd, KDSETMODE, KD_GRAPHICS );

	atexit(restore_vt);

	return 0;
	
}

int main(int argc, char **argv)
{
	int rc;

	log_file = fopen("ndrver.log", "a");

	LOG_INFO("Initializing IO spy\n");
	iospy_init();

	LOG_INFO("Initializing device tree...\n");

	if (seteuid(0) != 0 ){
                fprintf( stderr, "That tool must be run as root!\n");
                exit(1);
        }

	/* Initialize Device Tree access functions */
	rc = nrlib_initialize();
	if (rc) {
		LOG_ERR("Cannot initialize NameRegistryLib\n");
		return -1;
	}

	LOG_INFO("Initialize PCI access...\n");
	rc = ndrvpci_init_lib();
	if (rc) {
		LOG_ERR("Cannot initialize PCI\n");
		return -1;
	}

	/* Init low level asm trampoline */
	InitLinuxR2();

	/* Hack a mapping at 0 as some drivers still try to read the
	 * MacOS low memory globals
	 */
	mmap(0, 0x1000, PROT_READ /*| PROT_WRITE*/,
	     MAP_ANONYMOUS | MAP_FIXED | MAP_PRIVATE, 0, 0);

	LOG_INFO("Loading file based drivers...\n");

	/* Load file-based drivers */
	load_file_ndrvs();

#if 1
	LOG_INFO("Settign up VT...\n");
	if (setup_vt()) {
		LOG_ERR("VT setup failed\n");
		return 1;
	}
#endif
	LOG_INFO("Matching devices...\n");

	/* Scan device tree & match drivers */
	do_ndrv_matching();

	/* Open them all */
	do_ndrv_openall();

	restore_vt();
	
	/* Sleep them all */
	do_ndrv_suspendall();

	/* Wait for a key */
	LOG_INFO("Press RETURN...\n");
	(void)getchar();
	LOG_INFO("Resuming ...\n");

	/* Resume them all */
	do_ndrv_resumeall();

	fclose(log_file);

	return 0;
}
