#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <byteswap.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>

#include "ndrver.h"

#define PCIIOC_BASE		('P' << 24 | 'C' << 16 | 'I' << 8)
#define PCIIOC_CONTROLLER	(PCIIOC_BASE | 0x00)	/* Get controller for PCI device. */
#define PCIIOC_MMAP_IS_IO	(PCIIOC_BASE | 0x01)	/* Set mmap state to I/O space. */
#define PCIIOC_MMAP_IS_MEM	(PCIIOC_BASE | 0x02)	/* Set mmap state to MEM space. */
#define PCIIOC_WRITE_COMBINE	(PCIIOC_BASE | 0x03)	/* Enable/disable write-combining. */

struct ndrv_pci_inst
{
	struct device_node	*pci_node;
	int			domain, bus, dev, fn;
	int			procfd;
	unsigned long		configsave[0x30/4];
	int			restored;
	struct ndrv_pci_inst	*next;
};

static struct ndrv_pci_inst *ndrv_allpci;

int ndrvpci_init_lib(void)
{
	return 0;
}

static inline struct ndrv_pci_inst *node_to_pci(struct device_node *np)
{
	struct ndrv_pci_inst* pdata;

	for (pdata = ndrv_allpci; pdata; pdata = pdata->next) {
		struct device_node *p;

		for (p = np; p ; p = p->parent)
			if (p == pdata->pci_node)
				return pdata;
	}
	return NULL;
}

/* Walk up the device tree if necessary until we are on our enclosing
 * PCI device
 */
static struct device_node *find_pci_enclosure(struct device_node *np)
{
	struct device_node *pnp;

	pnp = np;
	while(pnp->parent) {
		if (!strcmp(pnp->parent->name, "pci") ||
		    !strcmp(pnp->parent->name, "pci-bridge"))
			break;
		pnp = pnp->parent;
	}
	if (pnp->parent == NULL)
		pnp = NULL;
	return pnp;
}

static int find_sysfs_pci_match(struct device_node *np, int *domain, int *bus,
				int *dev, int *fn)
{
	DIR *dir;
	struct dirent *entry;

	/* Iterate PCI devices in /sys/bus/pci/devices to find one which
	 * has a matching "devspec" entry
	 */
	
	dir = opendir("/sys/bus/pci/devices");
	if (dir == NULL)
		return -1;

	while ((entry = readdir(dir)) != NULL) {
		char path[1024], devspec[1024];
		int len, fd;

		if (!strcmp(entry->d_name, ".") ||
		    !strcmp(entry->d_name, ".."))
			continue;
		snprintf(path, 1024, "/sys/bus/pci/devices/%s/devspec", entry->d_name);
		fd = open(path, 0);
		if (fd == -1) {
			LOG_ERR("Cannot find %s\n", path);
			continue;
		}
		len = read(fd, devspec, 1023);
		devspec[len] = 0;
		close(fd);
		if (strcmp(devspec, np->full_name))
			continue;
		sscanf(entry->d_name, "%x:%x:%x.%x", domain, bus, dev, fn);
		break;
	}
	closedir(dir);
	return (entry != NULL) ? 0 : -1;
}

static int open_proc_bus_pci(struct ndrv_pci_inst *pcidata)
{
	char path[64];
	int i, aasize, acount, rc;
	unsigned int *assigned;
	unsigned int *aapladdr;

	// XXX FIXME: Fix domains, fix assumptions on assigned-addresses
	// format, etc...

	sprintf(path, "/proc/bus/pci/%02x/%02x.%x",
		pcidata->bus, pcidata->dev, pcidata->fn);
	pcidata->procfd = open(path, O_RDWR);
	if (pcidata->procfd == -1) {
		LOG_ERR("Failed to open %s\n", path);
		return -1;
	}

	LOG_INFO("Config space:\n");
	for (i=0; i<64; i+=4) {
		unsigned int data;

		lseek(pcidata->procfd, i, SEEK_SET);
		read(pcidata->procfd, &data, 4);
		LOG_INFO("%08x ", data);
		if ((i % 16) == 12)
			LOG_INFO("\n");
	}

	assigned = get_property(pcidata->pci_node, "assigned-addresses", &aasize);
	if (assigned == NULL) {
		LOG_ERR("Cannot find \"assigned-addresses\" property !\n");
		close(pcidata->procfd);
		return -1;
	}
	acount = aasize / (5 * 4);
	LOG_INFO("Found %d address entries\n", acount);
	aapladdr = malloc(sizeof(unsigned int) * acount);
	if (aapladdr == NULL) {
		LOG_ERR("Memory allocation failure for AAPL,address\n");
		close(pcidata->procfd);
		return -1;
	}
	for (i=0; i<acount; i++) {
		unsigned int physaddr, size;
		physaddr = assigned[5*i + 2];
		size = assigned[5*i + 4];
		LOG_INFO("Mapping 0x%08x (%d Kbytes)\n",
			 physaddr,  size/1024);
		rc = ioctl(pcidata->procfd, PCIIOC_MMAP_IS_MEM);
		if (rc)
			LOG_ERR("PCIIOC_MMAP_IS_MEM failed !\n");
		rc = ioctl(pcidata->procfd, PCIIOC_WRITE_COMBINE, 0); // XXX
		if (rc)
			LOG_ERR("PCIIOC_WRITE_COMBINE failed !\n");
		aapladdr[i] = (unsigned long)iospy_mmap(NULL, size, PROT_READ|PROT_WRITE,
							MAP_SHARED, pcidata->procfd, physaddr);
		if (aapladdr[i] == 0xffffffff/*(unsigned int)MAP_FAILED*/)
			LOG_ERR("mmap failed, err: %d\n", errno);

		// XXX Check failure
		LOG_INFO(" -> 0x%08x\n", aapladdr[i]);
	}
	rc = set_property(pcidata->pci_node, "AAPL,address", aapladdr,
			  sizeof(unsigned int)* acount);
	free(aapladdr);
	if (rc) {
		LOG_ERR("Can't set AAPL,address property\n");
		close(pcidata->procfd);
		return -1;
	}

	// XXX
	i = 0;
	rc = set_property(pcidata->pci_node, "AAPL,interrupts", &i, 4);

	return 0;
}

int ndrvpci_init_instance(struct ndrv_inst *ndrv)
{
	struct device_node *pnp;
	int domain, bus, dev, fn;
	struct ndrv_pci_inst *pcidata;

	pnp = find_pci_enclosure(ndrv->node);
	if (pnp == NULL) {
		LOG_ERR("Cannot match to a PCI bus\n");
		return -1;
	}
	LOG_INFO("Enclosing PCI device: %s\n", pnp->full_name);

	if (find_sysfs_pci_match(pnp, &domain, &bus, &dev, &fn)) {
		LOG_ERR("Cannot find match in sysfs !\n");
		return -1;
	}
	LOG_INFO("Matched to PCI device %04x:%02x:%02x.%1x\n",
		 domain, bus, dev, fn);

	for (pcidata = ndrv_allpci; pcidata; pcidata = pcidata->next)
		if (pcidata->domain == domain &&
		    pcidata->bus == bus &&
		    pcidata->dev == dev &&
		    pcidata->fn == fn) {
			LOG_INFO("Existing PCI instance !\n");
			ndrv->pcidata = pcidata;
			return 0;
		}

	pcidata = malloc(sizeof(struct ndrv_pci_inst));
	if (pcidata == NULL) {
		LOG_ERR("Allocation failure\n");
		return -1;
	}
	memset(pcidata, 0, sizeof(struct ndrv_pci_inst));

	pcidata->pci_node = pnp;
	pcidata->domain = domain;
	pcidata->bus = bus;
	pcidata->dev = dev;
	pcidata->fn = fn;

	if (open_proc_bus_pci(pcidata) != 0) {
		free(pcidata);
		return -1;
	}
	pcidata->next = ndrv_allpci;
	ndrv_allpci = pcidata;
	ndrv->pcidata = pcidata;

	return 0;
}

int ndrvpci_kernel_suspendall(void)
{
	struct ndrv_pci_inst *pdata;
	char syspath[1024];
	char state[16];
	int fd, len;

	for (pdata = ndrv_allpci; pdata; pdata = pdata->next) {
		sprintf(syspath, "/sys/bus/pci/devices/%04x:%02x:%02x.%x/power/state",
			pdata->domain, pdata->bus, pdata->dev, pdata->fn);
		fd = open(syspath, O_RDWR);
		if (fd == -1) {
			LOG_ERR("kernel suspend failed to open %s\n", syspath);
			continue;
		}
		/* Set special soft-suspend & lock state */
		len = sprintf(state, "10");
		write(fd, state, len);
		close(fd);
	}

	return 0;
}

int ndrvpci_kernel_resumeall(void)
{
	struct ndrv_pci_inst *pdata;
	char syspath[1024];
	char state[16];
	int fd, len;

	for (pdata = ndrv_allpci; pdata; pdata = pdata->next) {
		sprintf(syspath, "/sys/bus/pci/devices/%04x:%02x:%02x.%x/power/state",
			pdata->domain, pdata->bus, pdata->dev, pdata->fn);
		fd = open(syspath, O_RDWR);
		if (fd == -1) {
			LOG_ERR("kernel suspend failed to open %s\n", syspath);
			continue;
		}
		/* Set special unlock state */
		len = sprintf(state, "11");
		write(fd, state, len);
		close(fd);
		/* Set resume state */
		len = sprintf(state, "0");
		fd = open(syspath, O_RDWR);
		write(fd, state, len);
		close(fd);
	}

	return 0;
}

int ndrvpci_save_config(struct ndrv_inst *ndrv)
{
	int i, j = 0;

	for (i = 0x2c; i >= 0; i -= 4) {
		lseek(ndrv->pcidata->procfd, i, SEEK_SET);
		read(ndrv->pcidata->procfd, &ndrv->pcidata->configsave[j++], 4);
	}
	ndrv->pcidata->restored = 0;

	return 0;
}

int ndrvpci_restore_config(struct ndrv_inst *ndrv)
{
	int i, j = 0;

	if (ndrv->pcidata->restored)
		return 0;
	for (i = 0x2c; i >= 0; i -= 4) {
		lseek(ndrv->pcidata->procfd, i, SEEK_SET);	
		write(ndrv->pcidata->procfd, &ndrv->pcidata->configsave[j++], 4);
	}
	ndrv->pcidata->restored = 1;

	return 0;
}

UInt32 EndianSwap32Bit(UInt32 data)
{
	return bswap_32(data);
}

UInt16 EndianSwap16Bit(UInt16 data)
{
	return bswap_16(data);
}

OSErr ExpMgrConfigReadLong(RegEntryID * entryID, LogicalAddress offset, UInt32 * value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));
	UInt32 data;

	LOG_DBG("  ExpMgrConfigReadLong(%s, %lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset);

	lseek(pdata->procfd, (off_t)offset, SEEK_SET);
	read(pdata->procfd, &data, 4);
	*value = bswap_32(data);

	return noErr;
}

OSErr ExpMgrConfigWriteLong(RegEntryID * entryID, LogicalAddress offset, UInt32 value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));
	UInt32 data;

	LOG_DBG("  ExpMgrConfigWriteLong(%s, %lx, %08lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset, value);

	data = bswap_32(value);
	lseek(pdata->procfd, (off_t)offset, SEEK_SET);
	write(pdata->procfd, &data, 4);

	return noErr;
}

OSErr ExpMgrConfigReadWord(RegEntryID * entryID, LogicalAddress offset, UInt16 * value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));
	UInt16 data;

	LOG_DBG("  ExpMgrConfigReadWord(%s, %lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset);

	lseek(pdata->procfd, (off_t)offset, SEEK_SET);
	read(pdata->procfd, &data, 2);
	*value = bswap_16(data);

	return noErr;
}

OSErr ExpMgrConfigWriteWord(RegEntryID * entryID, LogicalAddress offset, UInt16 value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));
	UInt16 data;

	LOG_DBG("  ExpMgrConfigWriteWord(%s, %lx, %04x)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset, value);

	data = bswap_16(value);
	lseek(pdata->procfd, (off_t)offset, SEEK_SET);
	write(pdata->procfd, &data, 2);

	return noErr;
}

OSErr ExpMgrConfigReadByte(RegEntryID * entryID, LogicalAddress offset, UInt8 * value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrConfigReadByte(%s, %lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset);

	lseek(pdata->procfd, (off_t)offset, SEEK_SET);
	read(pdata->procfd, value, 1);

	return noErr;
}

OSErr ExpMgrConfigWriteByte(RegEntryID * entryID, LogicalAddress offset, UInt8 value)
{
       	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrConfigWriteByte(%s, %lx, %02x)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset, value);

	lseek(pdata->procfd, (off_t)offset, SEEK_SET);
	write(pdata->procfd, &value, 1);

	return noErr;
}

OSErr ExpMgrIOReadLong(RegEntryID * entryID, LogicalAddress offset, UInt32 * value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrIOReadLong(%s, %lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset);

	return -1;
}

OSErr ExpMgrIOWriteLong(RegEntryID * entryID, LogicalAddress offset, UInt32 value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrIOWriteLong(%s, %lx, %08lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset, value);

	return -1;
}

OSErr ExpMgrIOReadWord(RegEntryID * entryID, LogicalAddress offset, UInt16 * value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrIOReadWord(%s, %lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset);

	return -1;
}

OSErr ExpMgrIOWriteWord(RegEntryID * entryID, LogicalAddress offset, UInt16 value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrIOWriteWord(%s, %lx, %04x)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset, value);

	return -1;
}

OSErr ExpMgrIOReadByte(RegEntryID * entryID, LogicalAddress offset, UInt8 * value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrIOReadByte(%s, %lx)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset);

	return -1;
}

OSErr ExpMgrIOWriteByte(RegEntryID * entryID, LogicalAddress offset, UInt8 value)
{
	struct ndrv_pci_inst *pdata = node_to_pci(REGENTRY_TO_DN(entryID));

	LOG_DBG("  ExpMgrIOWriteByte(%s, %lx, %02x)\n",
		pdata ? pdata->pci_node->full_name : "<BAD>", (unsigned long)offset, value);

	return -1;
}

FunctionEntry PCILibFuncs[] =
    {
	MAKEFUNC( "ExpMgrConfigReadLong", ExpMgrConfigReadLong),
	MAKEFUNC( "ExpMgrConfigReadWord", ExpMgrConfigReadWord),
	MAKEFUNC( "ExpMgrConfigReadByte", ExpMgrConfigReadByte),
	MAKEFUNC( "ExpMgrConfigWriteLong", ExpMgrConfigWriteLong),
	MAKEFUNC( "ExpMgrConfigWriteWord", ExpMgrConfigWriteWord),
	MAKEFUNC( "ExpMgrConfigWriteByte", ExpMgrConfigWriteByte),

	MAKEFUNC( "ExpMgrIOReadLong", ExpMgrIOReadLong),
	MAKEFUNC( "ExpMgrIOReadWord", ExpMgrIOReadWord),
	MAKEFUNC( "ExpMgrIOReadByte", ExpMgrIOReadByte),
	MAKEFUNC( "ExpMgrIOWriteLong", ExpMgrIOWriteLong),
	MAKEFUNC( "ExpMgrIOWriteWord", ExpMgrIOWriteWord),
	MAKEFUNC( "ExpMgrIOWriteByte", ExpMgrIOWriteByte),

	MAKEFUNC( "EndianSwap16Bit", EndianSwap16Bit),
	MAKEFUNC( "EndianSwap32Bit", EndianSwap32Bit),
	ENDFUNC()
    };
