/*
 * Emulation for NameRegistryLib
 *
 * Note: We should add some more robustness using the remaining opaque
 * fields to store magic numbers and dealing with invalid entries,
 * though I yet have to find a driver passing us crap. Also, if we
 * ever implement interrupts as signals, check for re-entrancy and
 * bail out with proper error
 *
 * Also, iterators should be kept in a couple of lists so that they
 * are kept consistent accross property & node deletion.
 *
 * Finally, I'm pretty sure my implementation is semantically different
 * than Apple one in subtle ways. That's not a problem as long as
 * drivers work. No code is to be written for my semantics anyway.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <byteswap.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>

#include "ndrver.h"

//#define DBG(fmt...)	printf(fmt)
#define DBG(fmt...)


struct _RegEntryIter
{
	struct device_node *current;
	struct device_node *top;
	int op;
};

struct _RegPropertyIter
{
	struct device_node *node;
	struct property *prop;
};

struct device_node *dtroot, *allnodes, *macionode;

/*
 * Local implementation
 */

void *get_property(struct device_node *np, const char *name, int *length)
{
	struct property *prop = np->properties;

	while (prop) {
		if (!strcmp(name, prop->name)) {
			if (length)
				*length = prop->length;
			return prop->value;
		}
		prop = prop->next;
	}
	return NULL;
}


int set_property(struct device_node *np, const char *name, void *value, int size)
{
	struct property **pprop;
	struct property *prop = NULL;

	/* Delete existing */
	pprop = &np->properties;
	while ((prop = *pprop) != NULL) {
		if (strcmp(name, prop->name)) {
			pprop = &prop->next;
			continue;
		}
		if (prop->name)
			free(prop->name);
		if (prop->value)
			free(prop->value);
		*pprop = prop->next;
		break;
	}

	/* Add new */
	if (prop == NULL)
		prop = malloc(sizeof(struct property));
	if (prop == NULL)
		return -1;
	memset(prop, 0, sizeof(struct property));

	prop->name = malloc(strlen(name) + 1);
	if (prop->name == NULL) {
		free(prop);
		return -1;
	}
	strcpy(prop->name, name);

	prop->value = malloc(size);
	if (prop->value == NULL) {
		free(prop->name);
		free(prop);
		return -1;
	}
	memcpy(prop->value, value, size);
	prop->length = size;
	prop->next = np->properties;
	np->properties = prop;

	return 0;
}


int device_is_compatible(struct device_node *device, const char *compat)
{
	const char* cp;
	int cplen, l;

	cp = (char *) get_property(device, "compatible", &cplen);
	if (cp == NULL)
		return 0;
	while (cplen > 0) {
		if (strncasecmp(cp, compat, strlen(compat)) == 0)
			return 1;
		l = strlen(cp) + 1;
		cp += l;
		cplen -= l;
	}

	return 0;
}

static struct device_node *get_child(struct device_node *np, const char *name)
{
	struct device_node *c = np->child;

	while(c) {
		if (!strcmp(name, c->name))
			return c;
		c = c->sibling;
	}
	return NULL;
}

static int read_prop(char *realpath, struct property *prop)
{
	char proppath[1024];
	int fd, asize, rc;
	struct stat st;

	sprintf(proppath, "%s/%s", realpath, prop->name);
	fd = open(proppath, 0);
	if (fd == -1)
		return -1;
	
	prop->length = 0;
	rc = fstat(fd, &st);
	if (rc)
		fprintf(stderr, "cannot stat %s\n", proppath);
	else
		prop->length = st.st_size;

	/* For allocation of at least one byte even for an
	 * empty property
	 */
	asize = prop->length;
	if (asize == 0)
		asize = 1;

	prop->value = malloc(asize);
	memset(prop->value, 0, asize);
	if (prop->value == NULL) {
		close(fd);
		return -1;
	}
	read(fd, prop->value, prop->length);

	close(fd);

	return 0;
}

static struct device_node *read_node(char *curpath, struct device_node *parent)
{
	char realpath[1024];
	DIR *dir;
	struct dirent *entry;
	struct device_node *thisnode;
	struct device_node **childp;
	struct property *prop;
	int fplen;

	fplen = strlen(curpath);

	snprintf(realpath, 1024, "/proc/device-tree%s", curpath);

	DBG("* scanning: %s\n", realpath);

	dir = opendir(realpath);
	if (dir == NULL)
		return NULL;

	thisnode = malloc(sizeof(struct device_node));
	if (thisnode == NULL)
		goto bail;
	memset(thisnode, 0, sizeof(struct device_node));

	thisnode->full_name = malloc(strlen(curpath) + 1);
	if (thisnode->full_name == NULL) {
		free(thisnode); thisnode = NULL;
		goto bail;
	}
	sprintf(thisnode->full_name, "%s", curpath);
	thisnode->parent = parent;
	thisnode->allnext = allnodes;
	allnodes = thisnode;
	childp = &thisnode->child;
	
	while ((entry = readdir(dir)) != NULL) {
		switch(entry->d_type) {
		case DT_DIR:
			if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
				break;
			curpath[fplen] = 0;
			strcat(curpath, "/");
			strcat(curpath, entry->d_name);
			*childp = read_node(curpath, thisnode);
			if (*childp)
			    childp = &((*childp)->sibling);
			break;
		case DT_REG:
			prop = malloc(sizeof(struct property));
			if (prop == NULL)
				break;
			memset(prop, 0, sizeof(struct property));
			prop->name = malloc(strlen(entry->d_name) + 1);
			if (prop->name == NULL) {
				free(prop);
				break;
			}
			strcpy(prop->name, entry->d_name);
			if (read_prop(realpath, prop)) {
				free(prop->name);
				free(prop);
				break;
			}
			prop->next = thisnode->properties;
			thisnode->properties = prop;
			break;
		}
	}
	thisnode->name = get_property(thisnode, "name", NULL);
	if (thisnode->name == NULL)
		thisnode->name = "";

	/* Store first macio for use by path code (skip second macio on ws ?
	 * what about machines where it's called ohare/gc ?)
	 */
	if (macionode == NULL && !strcmp(thisnode->name, "mac-io"))
		macionode = thisnode;
 bail:
	curpath[fplen] = 0;
	closedir(dir);
	return thisnode;
}

int nrlib_initialize(void)
{
	char	curpath[1024];

	/* We make a local copy of the entire device tree now */
	curpath[0] = 0;
	dtroot = read_node(curpath, NULL);
	if (allnodes == NULL)
		return -1;

	return 0;
}

/*
 * Library entrypoints
 */

OSStatus RegistryEntryIDCopy(const RegEntryID * entryID, RegEntryID * to)
{
	memcpy(to, entryID, sizeof(RegEntryID));
	return noErr;
}

OSStatus RegistryEntryIDInit(RegEntryID * entryID)
{
	memset(entryID, 0, sizeof(*entryID));
	return noErr;
}

OSStatus RegistryEntryIDDispose(RegEntryID * entryID)
{
	SET_REGENTRY_TO_DN(entryID, NULL);

	return noErr;
}

OSStatus RegistryEntryDelete(const RegEntryID * entryID)
{
	struct device_node *np = REGENTRY_TO_DN(entryID);

	if (np == NULL)
		return -1;

	printf(" RegistryEntryDelete(%s) not implemented\n", np->full_name);

	return noErr;
}

Boolean RegistryEntryIDCompare(const RegEntryID * entryID1, const RegEntryID * entryID2)
{
	struct device_node *np1 = REGENTRY_TO_DN(entryID1);
	struct device_node *np2 = REGENTRY_TO_DN(entryID2);

	return np1 == np2;
}

OSStatus RegistryPropertyGetSize(const RegEntryID * entryID,
				 const RegPropertyName * propertyName,
				 RegPropertyValueSize * propertySize)
{
	struct device_node *np = REGENTRY_TO_DN(entryID);
	int size = 0;

	if (np == NULL)
		return nrNotFoundErr;

	LOG_DBG(" RegistryPropertyGetSize(%s, %s)\n", np->full_name, propertyName);

	void *prop = get_property(np, propertyName, &size);
	if (prop == NULL)
		return nrNotFoundErr;
	if (propertySize)
		*propertySize = size;

	return noErr;
}

OSStatus RegistryPropertyGet(const RegEntryID * entryID, 
			     const RegPropertyName * propertyName, void *propertyValue,
			     RegPropertyValueSize * propertySize)
{
	struct device_node *np = REGENTRY_TO_DN(entryID);
	int size = 0;

	if (np == NULL)
		return nrNotFoundErr;

	LOG_DBG(" RegistryPropertyGet(%s, %s)\n", np->full_name, propertyName);

	void *prop = get_property(np, propertyName, &size);
	if (prop == NULL)
		return nrNotFoundErr;
	if (propertySize == NULL)
		return -27;
	if (*propertySize < size)
		size = *propertySize;
	memcpy(propertyValue, prop, size);
	*propertySize = size;

	return noErr;
}

OSStatus RegistryPropertyCreate(const RegEntryID * entryID, const RegPropertyName * propertyName,
				const void * propertyValue, RegPropertyValueSize propertySize)
{
	struct device_node *np = REGENTRY_TO_DN(entryID);
	struct property *prop;

	if (np == NULL)
		return nrNotFoundErr;

	LOG_DBG(" RegistryPropertyCreate(%s, %s, %d bytes, ...)\n", np->full_name, propertyName,
		(int)propertySize);

	prop = np->properties;
	while(prop) {
		if (!strcmp(prop->name, propertyName))
			return nrPropertyAlreadyExists;
		prop = prop->next;
	}

	prop = malloc(sizeof(struct property));
	if (prop == NULL)
		return nrNotEnoughMemoryErr;
	memset(prop, 0, sizeof(struct property));

	prop->name = malloc(strlen(propertyName) + 1);
	if (prop->name == NULL) {
		free(prop);
		return nrNotEnoughMemoryErr;
	}
	strcpy(prop->name, propertyName);

	prop->value = malloc(propertySize);
	if (prop->value == NULL) {
		free(prop->name);
		free(prop);
		return nrNotEnoughMemoryErr;
	}
	memcpy(prop->value, propertyValue, propertySize);
	prop->length = propertySize;
	prop->next = np->properties;
	np->properties = prop;

	return noErr;
}

OSStatus RegistryPropertyDelete(const RegEntryID * entryID, const RegPropertyName * propertyName)
{
	struct device_node *np = REGENTRY_TO_DN(entryID);
	struct property *prop, **pprop;

	if (np == NULL)
		return nrNotFoundErr;

	LOG_DBG(" RegistryPropertyDelete(%s, %s)\n", np->full_name, propertyName);

	pprop = &np->properties;
	while ((prop = *pprop) != NULL) {
		if (strcmp(propertyName, prop->name)) {
			pprop = &prop->next;
			continue;
		}
		if (prop->name)
			free(prop->name);
		if (prop->value)
			free(prop->value);
		*pprop = prop->next;
		free(prop);
		return noErr;
	}

	return nrNotFoundErr;
}


OSStatus RegistryPropertySet(const RegEntryID * entryID, 
			     const RegPropertyName * propertyName, 
			     const void * propertyValue, RegPropertyValueSize propertySize)
{
	RegistryPropertyDelete(entryID, propertyName);
	return RegistryPropertyCreate(entryID, propertyName, propertyValue, propertySize);
}

OSStatus RegistryPropertyGetMod(const RegEntryID * entryID,
				const RegPropertyName * propertyName, 
				RegPropertyModifiers * mod)
{
	*mod = 0;
	return noErr;
}

OSStatus RegistryPropertySetMod(const RegEntryID * entryID,
				const RegPropertyName * propertyName,
				RegPropertyModifiers mod)
{
	return noErr;
}

OSStatus RegistryPropertyIterateCreate(const RegEntryID * entryID,
				       RegPropertyIter * cookie)
{
	struct device_node *np = REGENTRY_TO_DN(entryID);

	if (np == NULL)
		return -27;
	*cookie = malloc(sizeof(struct _RegPropertyIter));
	if (*cookie == NULL)
		return nrNotEnoughMemoryErr;
	(*cookie)->node = np;
	(*cookie)->prop = np->properties;

	return noErr;
}

OSStatus RegistryPropertyIterateDispose(RegPropertyIter * cookie)
{
	if (*cookie == NULL)
		return -27;
	free(*cookie);
	*cookie = NULL;
	return noErr;
}

OSStatus RegistryPropertyIterate(RegPropertyIter * cookie,
				 char * name, Boolean * done)
{
	struct property *prop;

	if (*cookie == NULL)
		return -27;

	/* -- Comment from Apple -- 
	 * Seems to be differences in handling "done".
	 * ATI assumes done = true when getting the last property.
	 * The Book says done is true after last property.
	 * ATI does check err, so this will work.
	 * Control ignores err and checks done.
	 */

	prop = (*cookie)->prop;
	if (prop == NULL)
		return nrIterationDone;
	prop = prop->next;
	*done = (prop == NULL);
	(*cookie)->prop = prop;

	return noErr;
}

OSStatus RegistryEntryIterateCreate(RegEntryIter * cookie)
{	
	*cookie = malloc(sizeof(struct _RegEntryIter));
	if (*cookie == NULL)
		return nrNotEnoughMemoryErr;
	memset(*cookie, 0, sizeof(struct _RegEntryIter));
	(*cookie)->op = kRegIterSubTrees;
	(*cookie)->current = dtroot;
	(*cookie)->top = dtroot;

	return noErr;
}

OSStatus RegistryEntryIterateDispose(RegEntryIter * cookie)
{
	if (*cookie == NULL)
		return -27;
	free(*cookie);
	*cookie = NULL;
	return noErr;
}

OSStatus RegistryEntryIterateSet(RegEntryIter * cookie,
				 const RegEntryID *startEntryID)
{
	struct device_node *np = REGENTRY_TO_DN(startEntryID);
	if (*cookie == NULL || np == NULL)
		return -27;
	(*cookie)->current = np;
	(*cookie)->top = dtroot;

	return noErr;
}

OSStatus RegistryEntryIterate(RegEntryIter * cookie,
			      UInt32 relationship,
			      RegEntryID * foundEntry,
			      Boolean *	done)
{
	struct device_node *np, *p;

	if (*cookie == NULL)
		return -27;
	np = (*cookie)->current;
	if (relationship != kRegIterContinue && relationship != (*cookie)->op)
		(*cookie)->op = relationship;
	/* Hrm, I'm not too sure about the exact meaning of the various
	 * relationships, though it seems Darwin doesn't care much, so...
	 */
	switch((*cookie)->op) {
	case kRegIterRoot:
		np = dtroot;
		(*cookie)->op = kRegIterSubTrees;
		break;
	case kRegIterParents:
		np = np->parent;
		break;
	case kRegIterChildren:
		np = np->child;
		(*cookie)->op = kRegIterSibling;
		break;
	case kRegIterSubTrees:
		if (np->child) {
			np = np->child;
			break;
		}
		if (np->sibling) {
			np = np->sibling;
			break;
		}
		while(1) {
			p = np->parent;
			if (p == NULL)
				break;
			if (p == (*cookie)->top) {
				np = NULL;
				break;
			}
			np = p->sibling;
			if (np)
				break;
		}
		break;
	case kRegIterSibling:
		np = np->sibling;
		break;
	}
	(*cookie)->current = np;
	SET_REGENTRY_TO_DN(foundEntry, np);	
	if (np == NULL) {
		*done = true;
		return nrNotFoundErr;
	}
	*done = false;
	return noErr;
}

OSStatus RegistryCStrEntryToName(const RegEntryID * entryID,
				 RegEntryID * parentEntry,
				 char * nameComponent,
				 Boolean * done )
{
	struct device_node *np = REGENTRY_TO_DN(entryID);

	if (np == NULL)
		return nrNotFoundErr;

	LOG_DBG("RegistryCStrEntryToName(%s)\n", np->full_name);

	strncpy(nameComponent, np->name, kRegMaximumPropertyNameLength);
	nameComponent[ kRegMaximumPropertyNameLength ] = 0;

	if (np->parent) {
		SET_REGENTRY_TO_DN(parentEntry, np->parent);
		*done = 0;
	} else
		*done = 1;
	return noErr;
}

OSStatus RegistryCStrEntryLookup(const RegEntryID * parentEntry,
				 const RegCStrPathName * path,
				 RegEntryID * newEntry)
{
	struct device_node *parent = REGENTRY_TO_DN(parentEntry);
	struct device_node *np = NULL;
	char tmpbuf[1024];
	char *p;

#define kDTRoot		"Devices:device-tree:"
#define kMacIORoot	"Devices:device-tree:pci:mac-io:"

	if (path[0] == ':')
		path++;
	else if (!strncmp(path, kDTRoot, strlen(kDTRoot))) {
		path += strlen(kDTRoot);
		parent = dtroot;
	} else if (!strncmp(path, kMacIORoot, strlen(kMacIORoot))) {
		path += strlen(kMacIORoot);
		parent = macionode;
	}
	p = tmpbuf;
	while(*path) {
		if (*path == ':') {
			*p = 0;
			parent = get_child(parent, tmpbuf);
			p = tmpbuf;
		}
		path++;
	}
	if (p == tmpbuf)
		np = parent;
	else {
		*p = 0;
		np = get_child(parent, tmpbuf);
	}
	if (np == NULL)
		return nrNotFoundErr;
	SET_REGENTRY_TO_DN(newEntry, np);
	
	return noErr;
}


OSStatus RegistryCStrEntryCreate(const RegEntryID * parentEntry,
				 const RegCStrPathName * name,
				 RegEntryID * newEntry)
{
	struct device_node *parent = REGENTRY_TO_DN(parentEntry);
	struct device_node *np, **linkp;
	char tmpbuf[1024];
	int len;

	if (parent == NULL)
		parent = dtroot;

	linkp = &parent->child;
	while(*linkp)
		linkp = &((*linkp)->sibling);

	np = malloc(sizeof(struct device_node));
	if (np == NULL)
		return nrNotEnoughMemoryErr;
	memset(np, 0, sizeof(struct device_node));

	len = sprintf(tmpbuf, "%s/%s", parent->full_name, name);
	np->full_name = malloc(len + 1);
	if (np->full_name == NULL) {
		free(np);
		return nrNotEnoughMemoryErr;
	}
	strcpy(np->full_name, tmpbuf);

	SET_REGENTRY_TO_DN(newEntry, np);
	if (RegistryPropertyCreate(newEntry, "name", name, strlen(name)) != noErr) {
		free(np->full_name);
		free(np);
		SET_REGENTRY_TO_DN(newEntry, NULL);
		return nrNotCreatedErr;
	}
	np->parent = parent;
	np->allnext = allnodes;
	allnodes = np;
	*linkp = np;
	
	return noErr;
}

FunctionEntry NameRegistryLibFuncs[] = {
	MAKEFUNC( "RegistryEntryIDCopy", RegistryEntryIDCopy),
	MAKEFUNC( "RegistryEntryIDInit", RegistryEntryIDInit),
	MAKEFUNC( "RegistryEntryIDDispose", RegistryEntryIDDispose),
	MAKEFUNC( "RegistryEntryIDCompare", RegistryEntryIDCompare),
	MAKEFUNC( "RegistryPropertyGetSize", RegistryPropertyGetSize),
	MAKEFUNC( "RegistryPropertyGet", RegistryPropertyGet),
	MAKEFUNC( "RegistryPropertyGetMod", RegistryPropertyGetMod),
	MAKEFUNC( "RegistryPropertySetMod", RegistryPropertySetMod),

	MAKEFUNC( "RegistryPropertyIterateCreate", RegistryPropertyIterateCreate),
	MAKEFUNC( "RegistryPropertyIterateDispose", RegistryPropertyIterateDispose),
	MAKEFUNC( "RegistryPropertyIterate", RegistryPropertyIterate),

	MAKEFUNC( "RegistryEntryIterateCreate", RegistryEntryIterateCreate),
	MAKEFUNC( "RegistryEntryIterateDispose", RegistryEntryIterateDispose),
	MAKEFUNC( "RegistryEntryIterate", RegistryEntryIterate),
	MAKEFUNC( "RegistryEntryIterateSet", RegistryEntryIterateSet),
	MAKEFUNC( "RegistryCStrEntryToName", RegistryCStrEntryToName),
	MAKEFUNC( "RegistryCStrEntryLookup", RegistryCStrEntryLookup),

	MAKEFUNC( "RegistryCStrEntryCreate", RegistryCStrEntryCreate),
	MAKEFUNC( "RegistryEntryDelete", RegistryEntryDelete),

	MAKEFUNC( "RegistryPropertyCreate", RegistryPropertyCreate),
	MAKEFUNC( "RegistryPropertyDelete", RegistryPropertyDelete),
	MAKEFUNC( "RegistryPropertySet", RegistryPropertySet),
	ENDFUNC(),
};


